import logging
from typing import Optional, Callable
from .net.package import Package
from .net.protocol import Protocol


class ApiProtocol(Protocol):

    PROTO_REQ_DATA = 0x00
    PROTO_REQ_ALERTS_COUNT = 0x02

    PROTO_RES_ALERTS_COUNT = 0x81
    PROTO_RES_ERR = 0xe0
    PROTO_RES_OK = 0xe1

    def __init__(self, connection_lost: Callable):
        super().__init__()
        self.set_connection_lost(connection_lost)

    def connection_lost(self, exc: Optional[Exception]):
        super().connection_lost(exc)
        self._connection_lost()

    def set_connection_lost(self, connection_lost: Callable):
        self._connection_lost = connection_lost

    def _on_res_data(self, pkg):
        future = self._get_future(pkg)
        if future is None:
            return
        future.set_result(pkg.data)

    def _on_res_err(self, pkg: Package):
        future = self._get_future(pkg)
        if future is None:
            return
        future.set_exception(Exception(pkg.data))

    def _on_res_ok(self, pkg):
        future = self._get_future(pkg)
        if future is None:
            return
        future.set_result(None)

    def on_package_received(self, pkg, _map={
        PROTO_RES_ALERTS_COUNT: _on_res_data,
        PROTO_RES_ERR: _on_res_err,
        PROTO_RES_OK: _on_res_ok,
    }):
        handle = _map.get(pkg.tp)
        if handle is None:
            logging.error(f'Unhandled package type: {pkg.tp}')
        else:
            handle(self, pkg)



import asyncio
import logging
from typing import Union, Optional
from .package import Package


class Protocol(asyncio.Protocol):

    _connected = False

    def __init__(self):
        super().__init__()
        self._buffered_data = bytearray()
        self._package = None
        self._requests = dict()
        self._pid = 0
        self.transport = None

    def connection_made(self, transport: asyncio.BaseTransport):
        '''
        override asyncio.Protocol
        '''
        self.transport = transport

    def connection_lost(self, exc: Optional[Exception]):
        '''
        override asyncio.Protocol
        '''
        self.transport = None
        self._package = None
        self._buffered_data.clear()

    def is_connected(self) -> bool:
        return self.transport is not None

    def request(self, pkg: Package,
                timeout: Union[None, float, int] = None
                ) -> asyncio.Future:
        self._pid += 1
        self._pid %= 0x10000

        pkg.pid = self._pid

        task = asyncio.ensure_future(
            self._timer(pkg.pid, timeout)) if timeout else None

        future = asyncio.Future()
        self._requests[pkg.pid] = (future, task)

        self.transport.write(pkg.to_bytes())

        return future

    def data_received(self, data: bytes):
        '''
        override asyncio.Protocol
        '''
        self._buffered_data.extend(data)
        while self._buffered_data:
            size = len(self._buffered_data)
            if self._package is None:
                if size < Package.st_package.size:
                    return None
                self._package = Package(self._buffered_data)
            if size < self._package.total:
                return None
            try:
                self._package.extract_data_from(self._buffered_data)
            except KeyError as e:
                logging.error('Unsupported package received: {}'.format(e))
            except Exception as e:
                logging.exception(e)
                # empty the byte-array to recover from this error
                self._buffered_data.clear()
            else:
                self.on_package_received(self._package)
            self._package = None

    def on_package_received(self, pkg: Package):
        raise NotImplementedError

    async def _timer(self, pid: int, timeout: Union[float, int]):
        await asyncio.sleep(timeout)
        try:
            future, task = self._requests.pop(pid)
        except KeyError:
            logging.error('Timed out package id not found: {}'.format(
                self._package.pid))
            return None

        future.set_exception(TimeoutError(
            'request timed out on package id {}'.format(pid)))

    def _get_future(self, pkg: Package) -> asyncio.Future:
        future, task = self._requests.pop(pkg.pid, (None, None))
        if future is None:
            logging.error(
                f'Got a response on pkg id {pkg.pid} but the original '
                'request has probably timed-out'
            )
            return
        task.cancel()
        return future



import abc
from .asset import Asset
from typing import Tuple, List, Optional


class CheckBase(abc.ABC):
    key: str  # Check key (must not be changed)

    def __init_subclass__(cls, **kwargs):
        if not hasattr(cls, 'key'):
            raise NotImplementedError('key not implemented')
        if not isinstance(cls.key, str):
            raise NotImplementedError('key must be type str')
        return super().__init_subclass__(**kwargs)

    @abc.abstractclassmethod
    async def run(cls, ts: float, asset: Asset) -> Tuple[
            Optional[dict], Optional[dict]]:
        ...


class CheckBaseMulti(abc.ABC):
    key: str  # Check key (must not be changed)

    def __init_subclass__(cls, **kwargs):
        if not hasattr(cls, 'key'):
            raise NotImplementedError('key not implemented')
        if not isinstance(cls.key, str):
            raise NotImplementedError('key must be type str')
        return super().__init_subclass__(**kwargs)

    @abc.abstractclassmethod
    async def run(cls, ts: float, assets: List[Asset]) -> List[
            Tuple[Optional[dict], Optional[dict]]]:
        ...



import asyncio
import logging
import random
from typing import Optional, Dict, List, Awaitable
from .net.package import Package
from .protocol import ApiProtocol


class HubClient:

    def __init__(self, auto_reconnect: bool = True):
        self._loop = asyncio.get_event_loop()
        self._protocol = None
        self._reconnect = auto_reconnect
        self._reconnecting = False
        self._pool = None
        self._pool_idx = 0
        self._request = self._ensure_write if auto_reconnect else self._write

    def close(self):
        if self._protocol and self._protocol.transport:
            self._protocol.transport.close()
        self._protocol = None

    def is_connected(self) -> bool:
        return self._protocol is not None and self._protocol.is_connected()

    def connect_pool(self, pool: list) -> asyncio.Future:
        assert self.is_connected() is False
        assert self._reconnecting is False
        assert len(pool), 'pool must contain at least one node'

        self._pool = tuple((
            (address, 8700) if isinstance(address, str) else address
            for address in pool))
        self._pool_idx = random.randint(0, len(pool) - 1)
        return self.reconnect()

    def connect(self, host: str, port: int = 8700,
                timeout: Optional[int] = 5) -> asyncio.Future:
        assert self.is_connected() is False
        self._pool = ((host, port),)
        self._pool_idx = 0
        return self._connect(timeout=timeout)

    def reconnect(self) -> Optional[asyncio.Future]:
        if self._reconnecting:
            return
        self._reconnecting = True
        return asyncio.ensure_future(self._reconnect_loop())

    async def _connect(self, timeout=5):
        host, port = self._pool[self._pool_idx]
        try:
            conn = self._loop.create_connection(
                lambda: ApiProtocol(self._on_connection_lost),
                host=host,
                port=port
            )
            _, self._protocol = await asyncio.wait_for(
                conn,
                timeout=timeout)
        finally:
            self._pool_idx += 1
            self._pool_idx %= len(self._pool)

    def _on_connection_lost(self):
        self._protocol = None
        if self._reconnect:
            self.reconnect()

    async def _reconnect_loop(self):
        try:
            wait_time = 1
            timeout = 2
            protocol = self._protocol
            while True:
                host, port = self._pool[self._pool_idx]
                try:
                    await self._connect(timeout=timeout)
                except Exception as e:
                    logging.error(
                        f'Connecting to {host}:{port} failed: '
                        f'{e}({e.__class__.__name__}), '
                        f'Try next connect in {wait_time} seconds'
                    )
                else:
                    if protocol and protocol.transport:
                        self._loop.call_later(10.0, protocol.transport.close)
                    break

                await asyncio.sleep(wait_time)
                wait_time *= 2
                wait_time = min(wait_time, 60)
                timeout = min(timeout + 1, 10)
        finally:
            self._reconnecting = False

    async def _ensure_write(self, pkg):
        if not self._pool:
            raise ConnectionError('no connection')
        while True:
            if not self.is_connected():
                logging.info('Wait for a connection')
                self.reconnect()  # ensure the re-connect loop
                await asyncio.sleep(1.0)
                continue

            try:
                res = await self._protocol.request(pkg, timeout=10)
            except Exception as e:
                logging.error(
                    f'Failed to transmit package: '
                    f'{e}({e.__class__.__name__}) (will try again)')
                await asyncio.sleep(1.0)
                continue

            return res

    async def _write(self, pkg):
        if not self.is_connected():
            raise ConnectionError('no connection')
        return await self._protocol.request(pkg, timeout=10)

    def send_check_data(self, path: list, check_data: dict) -> Awaitable:
        pkg = Package.make(
            ApiProtocol.PROTO_REQ_DATA,
            data=[path, check_data],
            partid=path[0],  # asset_id
        )
        return self._request(pkg)

    def get_alerts_count(self, container_ids: list,
                         asset_ids: Optional[list] = None,
                         user_id: Optional[int] = None) -> Awaitable[list]:
        pkg = Package.make(
            ApiProtocol.PROTO_REQ_ALERTS_COUNT,
            data=[container_ids, asset_ids, user_id]
        )
        return self._request(pkg)



import asyncio
import logging
import signal
import os
from asyncio import AbstractEventLoop
from typing import Tuple, Union, Optional, Callable
from .check import CheckBase, CheckBaseMulti
from .serviceroom import ServiceRoom
from .hub import hub
from .logger import setup_logger, set_log_level
from .ticonn import ticonn


HUB_HOST = os.getenv('HUB_HOST', 'hub')
HUB_PORT = int(os.getenv('HUB_PORT', '8700'))
THINGSDB_HOSTLIST = os.getenv('THINGSDB_HOSTLIST', 'thingsdb:9200')
THINGSDB_TOKEN = os.getenv('THINGSDB_TOKEN')
THINGSDB_SCOPE = os.getenv('THINGSDB_SCOPE', '//data')


async def _setup_ticonn():
    nodes = [
        tuple(node.split(':'))
        for node in THINGSDB_HOSTLIST.replace(';', ',').split(',')]
    token = THINGSDB_TOKEN
    ticonn.set_default_scope(THINGSDB_SCOPE)
    await ticonn.connect_pool(nodes, token)
    logging.info(f'Connected to ThingsDB (ticonn)')


async def _setup_hub_connection():
    logging.info('Connecting to Hub')
    await hub.connect(HUB_HOST, HUB_PORT)


async def _setup_svc_room(collector_key, checks, no_count, max_timeout):
    service_room = ServiceRoom('.ev_service.id()', THINGSDB_SCOPE)
    service_room.init(
        collector_key,
        checks,
        set_log_level,
        no_count,
        max_timeout)
    await service_room.join(ticonn)
    await service_room.load_all()
    asyncio.ensure_future(service_room.run_loop())
    logging.info(f'Collection `{THINGSDB_SCOPE}` ready')


def _stop(signame, *args):
    logging.warning(f'Signal \'{signame}\' received, stop service')
    for task in asyncio.all_tasks():
        task.cancel()
    raise Exception


def start(collector_key: str, version: str,
          checks: Tuple[Union[CheckBase, CheckBaseMulti]],
          start_func: Optional[Callable[[AbstractEventLoop], None]] = None,
          close_func: Optional[Callable[[AbstractEventLoop], None]] = None,
          no_count: bool = False, max_timeout: float = 300.0):
    if THINGSDB_TOKEN is None:
        raise Exception('Missing `THINGSDB_TOKEN` environment variable')

    setup_logger()
    logging.warning(f"Starting {collector_key} service v{version}")

    loop = asyncio.get_event_loop()

    if start_func is not None:
        start_func(loop)

    loop.run_until_complete(_setup_ticonn())
    loop.run_until_complete(_setup_hub_connection())
    loop.run_until_complete(_setup_svc_room(
        collector_key,
        checks,
        no_count,
        max_timeout))

    signal.signal(signal.SIGINT, _stop)
    signal.signal(signal.SIGTERM, _stop)

    try:
        loop.run_forever()
    except Exception:
        loop.run_until_complete(loop.shutdown_asyncgens())

    ticonn.close()
    loop.run_until_complete(ticonn.wait_closed())

    if close_func is not None:
        close_func(loop)

    logging.info('Bye.')



import base64
from asyncio import Lock
from typing import Optional
from cryptography.fernet import Fernet
from .ticonn import ticonn


class Asset:
    __slots__ = (
        'container_id', 'asset_id', 'check_id', 'config', 'key', '_lock')

    container_id: int
    asset_id: int
    check_id: int
    config: dict
    key: Optional[str]
    lock: Optional[Lock]

    def __init__(self, container_id: int, asset_id: int, check_id: int,
                 config: dict):
        self.container_id = container_id
        self.asset_id = asset_id
        self.check_id = check_id
        self.config = config
        self.key = None
        self._lock = None

    def __repr__(self) -> str:
        return f"asset: {self.asset_id} check: {self.check_id}"

    def get_interval(self) -> int:
        """Returns the check interval in seconds."""
        return self.config['_interval']

    async def decrypt(self, secret: str):
        """Returns the decrypted value."""
        if self.key is None:
            self.key = await ticonn.run(
                'get_encryption_key',
                self.container_id)

        return Fernet(self.key).decrypt(base64.b64decode(secret)).decode()

    def get_lock(self) -> Lock:
        if self._lock is None:
            self._lock = Lock()
        return self._lock



from .severity import Severity
from typing import Optional


class CheckException(Exception):
    """CheckException is the basic check exception."""
    def __init__(self, msg: str, severity: Severity = Severity.MEDIUM):
        assert msg, 'CheckException message must not be empty'
        self.severity = severity
        super().__init__(msg)

    def to_dict(self):
        return {
            "error": self.__str__(),
            "severity": self.severity.value
        }


class NoCountException(Exception):  # Only to be used within CheckBase
    """NoCountException must ONLY be raised withing non-multi CheckBase class!!
    This exception cannot be used with CheckBaseMulti;
    """
    def __init__(self, result: dict):
        assert isinstance(result, dict)
        self.result = result
        super().__init__('No count exception')



from __future__ import annotations
import logging
import msgpack
import struct
from typing import Any, Optional


class Package(object):

    __slots__ = ('partid', 'length', 'pid', 'tp', 'body', 'data', 'total')

    st_package = struct.Struct('<QIHBB')

    def __init__(self, barray: Optional[bytearray] = None):
        if barray is None:
            return

        self.partid, self.length, self.pid, self.tp, checkbit = \
            self.__class__.st_package.unpack_from(barray, offset=0)
        if self.tp != checkbit ^ 255:
            raise ValueError('invalid checkbit')
        self.total = self.__class__.st_package.size + self.length
        self.body = None
        self.data = None

    @classmethod
    def make(
        cls,
        tp: int,
        data: bytes = b'',
        pid: int = 0,
        partid: int = 0,
        is_binary: bool = False,
    ) -> Package:
        pkg = cls()
        pkg.tp = tp
        pkg.pid = pid
        pkg.partid = partid

        if is_binary is False:
            data = msgpack.packb(data)

        pkg.body = data
        pkg.length = len(data)
        return pkg

    def to_bytes(self) -> bytes:
        header = self.st_package.pack(
            self.partid,
            self.length,
            self.pid,
            self.tp,
            self.tp ^ 0xff)

        return header + self.body

    def extract_data_from(self, barray: bytearray):
        self.body = None
        try:
            if self.length:
                self.body = barray[self.__class__.st_package.size:self.total]
                self.data = msgpack.unpackb(
                    self.body,
                    use_list=False,
                    strict_map_key=False)
        finally:
            del barray[:self.total]

    def read_data(self) -> Any:
        if self.data:
            return self.data
        if self.body is None:
            return
        try:
            self.data = msgpack.unpackb(self.body)
        except Exception:
            logging.error('Failed to unpack package id: {0.pid}'.format(self))
            raise
        return self.data

    def __repr__(self) -> str:
        return '<id: {0.pid} size: {0.length} tp: {0.tp}>'.format(self)



import asyncio
import functools
import logging
import time
import os
import json
import sys
from collections import defaultdict
from thingsdb.room import Room
from thingsdb.room import event
from typing import List, Dict, Tuple, Callable, Union, Optional
from .hub import hub
from .exceptions import CheckException, NoCountException
from .asset import Asset
from .check import CheckBase, CheckBaseMulti


HUB_REQ_SLEEP = .001
SLEEP_TIME = int(os.getenv('SLEEP_TIME', 2))
DRY_RUN = os.getenv('DRY_RUN', '0') != '0'
assert 60 >= SLEEP_TIME > 0


class ServiceRoom(Room):

    def init(self, collector_key: str,
             checks: Tuple[Union[CheckBase, CheckBaseMulti]],
             on_log_level: Callable[[int], None], no_count: bool = False,
             max_timeout: float = 300.0):
        self.collector_key = collector_key
        self._checks = {check.key: check for check in checks}
        self._last = int(time.time())-1
        self._scheduled: Dict[int, Dict[int, dict]] = defaultdict(dict)
        self._on_log_level = on_log_level
        self._no_count = no_count
        self._max_timeout = max_timeout

    async def load_all(self):
        self._query = functools.partial(self.client.query, scope=self.scope)
        assert self.collector_key, 'run init before load_all()'
        root_id, root = await self._query("""//ti
            [.root.id(), .root.wrap("_ContainerSvc")];
        """)
        await self._on_container(root_id, root)

    def _get_work(self):
        now = int(time.time())
        if now <= self._last:
            logging.warning('now before or equal to last; '
                            'maybe the clock time has changed?')
            return {}
        work = defaultdict(list)
        for (container_id, asset_id), checks in self._scheduled.items():
            for check_id, (check_key, config) in checks.items():
                interval = config['_interval']
                diff = now - self._last
                part = (asset_id % interval + now) % interval
                if part < diff:
                    work[check_key].append(Asset(
                        container_id,
                        asset_id,
                        check_id,
                        config))
        self._last = now
        return work

    async def _load(self, container_id: int):
        container = await self._query("""//ti
            Container(id).wrap("_ContainerSvc");
        """, id=container_id)
        await self._on_container(container_id, container)

    async def _on_container(self, container_id: int, container: dict):
        assets = container['assets']
        children = container['children']
        for (asset_id, services) in assets:
            if not services:
                continue
            key = (container_id, asset_id)
            for (collector_key, svc_config, checks) in services:
                if collector_key != self.collector_key:
                    continue
                for check_id, check_key, interval, chk_config in checks:
                    config = {
                        '_interval': interval,
                        **(svc_config or {}),  # can be empty
                        **(chk_config or {}),  # can be empty
                    }
                    self._scheduled[key][check_id] = (check_key, config)

        for cid in children:
            await self._load(cid)

    async def _send_to_hub(self, asset: Asset, result: Optional[dict],
                           error: Optional[dict], ts: float, no_count: bool):
        if error:
            logging.error(f'Error: {error}; {asset}')

        path = asset.asset_id, asset.check_id
        check_data = {
            'result': result,
            'error': error,
            'framework': {
                'duration': time.time() - ts,
                'timestamp': int(ts),
                'no_count': no_count,
            }
        }
        try:
            if DRY_RUN:
                output = json.dumps(check_data, indent=2)
                print('-'*80, file=sys.stderr)
                print(output)
                print('', file=sys.stderr)
            else:
                await hub.send_check_data(path, check_data)
        except Exception as e:
            msg = str(e) or type(e).__name__
            logging.error(f'Failed to send data to hub: {msg}; {asset}')
        else:
            logging.debug(f'Successfully send data to hub; {asset}')

    async def _run_multi(self, check: CheckBaseMulti, assets: List[Asset]):
        ts = time.time()
        try:
            results = await asyncio.wait_for(
                check.run(ts, assets),
                timeout=self._max_timeout)
        except asyncio.TimeoutError:
            error = CheckException('timed out').to_dict()
            results = [(None, error)] * len(assets)
        except CheckException as e:
            error = e.to_dict()
            results = [(None, error)] * len(assets)
        except Exception as e:
            msg = str(e) or type(e).__name__
            error = CheckException(msg).to_dict()
            results = [(None, error)] * len(assets)

        for asset, (result, error) in zip(assets, results):
            await self._send_to_hub(asset, result, error, ts, self._no_count)
            await asyncio.sleep(HUB_REQ_SLEEP)

    async def _run(self, check: CheckBase, asset: Asset):
        ts = time.time()
        no_count = self._no_count
        timeout = min(0.8 * asset.get_interval(), self._max_timeout)
        try:
            result, error = await asyncio.wait_for(
                check.run(ts, asset),
                timeout=timeout)
        except asyncio.TimeoutError:
            result, error = None, CheckException('timed out').to_dict()
        except NoCountException as e:
            result, error, no_count = e.result, None, True  # Force True
        except CheckException as e:
            result, error = None, e.to_dict()
        except Exception as e:
            msg = str(e) or type(e).__name__
            result, error = None, CheckException(msg).to_dict()

        await self._send_to_hub(asset, result, error, ts, no_count)

    async def run_loop(self):
        while True:
            work = self._get_work()
            total = len(self._scheduled)
            log_msg = f'Work: {len(work)} item(s), total: {total}'
            if work:
                logging.info(log_msg)
            else:
                logging.debug(log_msg)
            for check_key, assets in work.items():
                check = self._checks.get(check_key)
                if check is None:
                    logging.warning(f'Check {check_key} not implemented')
                    continue

                if issubclass(check, CheckBaseMulti):
                    asyncio.ensure_future(self._run_multi(check, assets))
                else:
                    for asset in assets:
                        asyncio.ensure_future(self._run(check, asset))
            for _ in range(SLEEP_TIME):
                await asyncio.sleep(1)

    @event('set-log-level')
    def on_set_log_level(self, log_level: str):
        self._on_log_level(log_level)

    @event('upsert-asset')
    def on_upsert_asset(self, container_id: int,
                        service_data: Tuple[int, tuple]):
        logging.debug('on upsert asset')
        asset_id, services = service_data
        self._scheduled.pop(asset_id, None)
        key = (container_id, asset_id)
        self._scheduled.pop(key, None)

        for (collector_key, svc_config, checks) in services:
            if collector_key != self.collector_key:
                continue

            for check_id, check_key, interval, chk_config in checks:
                config = {
                    '_interval': interval,
                    **(svc_config or {}),  # can be empty
                    **(chk_config or {}),  # can be empty
                }
                self._scheduled[key][check_id] = (check_key, config)

    @event('unset-assets')
    def on_unset_assets(self, container_id: int, asset_ids: Tuple[int]):
        logging.debug('on unset assets')
        for asset_id in asset_ids:
            key = (container_id, asset_id)
            self._scheduled.pop(key, None)
